#!/usr/bin/python

import math

import pygtk
pygtk.require("2.0")
import gtk

import cairo

import rsvg

import threading
from Queue import Queue, Empty

import time


TOOL_SELECT = 1
TOOL_LINE = 2
TOOL_RECTANGLE = 3
TOOL_ELLIPSE = 4
TOOL_TEXT = 5
TOOL_SMILE = 6

class DrawElement:
    def __init__(self, x, y, fg, bg):
        self.x, self.y = x, y
        self.w, self.h = 1, 1
        self.aw, self.ah = 1, 1
        self.bg, self.fg = bg, fg
    
    def update(self, w, h):
        w = max(1., w)
        h = max(1., h)
        
        self.aw = max(w, self.aw)
        self.ah = max(h, self.ah)
        
        self.w, self.h = w, h
    
    def inside(self, x, y):
        return self.x < x < (self.x + self.w) and \
               self.y < y < (self.y + self.h)
    
    def get_area(self):
        area = gtk.gdk.Rectangle(int(self.x) - 4, int(self.y) - 4, 
                                 int(self.aw) + 8, int(self.ah) + 8)
        return area
    
    def draw(self, cr):
        pass

class Line(DrawElement):    
    def draw(self, cr):
        old_dash, old_offset = cr.get_dash()
        
        cr.set_dash(10 * [3], 0.0)
        
        cr.save()
        cr.translate(self.x, self.y)
        
        cr.move_to(0.0, 0.0)
        cr.line_to(self.w, self.h)
        cr.set_source_rgb(*self.fg)
        cr.stroke()
        
        cr.restore()
        
        cr.set_dash(old_dash, old_offset)

class Rectange(DrawElement):    
    def draw(self, cr):
        grad = cairo.LinearGradient(0, 0, 1, 1)
        grad.add_color_stop_rgb(0, *self.fg)
        grad.add_color_stop_rgb(1.0, *self.bg)
        
        cr.save()
        
        cr.translate(self.x, self.y)
        cr.scale(self.w, self.h)
        
        cr.rectangle(0., 0., 1., 1.)
        cr.set_line_width(max(cr.device_to_user_distance(3, 3)))
        cr.set_source_rgb(*self.fg)
        cr.stroke()
        
        cr.rectangle(0., 0., 1., 1.)
        cr.set_source(grad)
        cr.fill()
        
        cr.restore()
        
class Ellipse(DrawElement):   
    def draw(self, cr):
        cr.save()
        
        cr.translate(self.x + self.w / 2., self.y + self.w / 2.)
        cr.scale(self.w / 2., self.h / 2.)
        
        cr.arc(0., 0., 1., 0., 2 * math.pi)    
        cr.set_line_width(max(cr.device_to_user_distance(3, 3)))
        cr.set_source_rgb(*self.fg)
        cr.stroke()
        
        cr.arc(0., 0., 1., 0., 2 * math.pi)
        cr.set_source_rgb(*self.bg)
        cr.fill()
        
        cr.restore()

class Text(DrawElement):
    def draw(self, cr):
        cr.save()
        
        cr.translate(self.x, self.y)
        
        cr.set_source_rgb(*self.fg) 
        cr.move_to(0.0, self.h)
        cr.set_font_size(self.h)
        cr.show_text('HELLO, WORLD!')
        
        cr.restore()

class Smile(DrawElement):
    svg = None
    
    def __init__(self, *args):
        DrawElement.__init__(self, *args)
        
        if Smile.svg is None:
            Smile.svg = rsvg.Handle(file = 'smiley.svg')
    
    def draw(self, cr):
        cr.save()
        
        cr.translate(self.x, self.y)
        
        sw, sh = Smile.svg.get_dimension_data()[2:]
        
        cr.scale(self.w / sw, self.h / sh)
        
        Smile.svg.render_cairo(cr)
        
        cr.restore()

class MyDrawingArea(gtk.DrawingArea):
    def __init__(self):
        gtk.DrawingArea.__init__(self)
        
        self.bg_color = (1., 1., 1.)
        self.fg_color = (0., 0., 0.)
        
        self.tool = TOOL_SELECT
        
        self.add_events(gtk.gdk.BUTTON_PRESS_MASK | 
                        gtk.gdk.BUTTON_RELEASE_MASK |  
                        gtk.gdk.POINTER_MOTION_MASK | 
                        gtk.gdk.SCROLL_MASK)
        
        self.connect("expose_event", self.expose)
        
        self.connect("button_press_event", self.on_button_press)
        self.connect("button_release_event", self.on_button_release)
        self.connect("motion_notify_event", self.on_motion)
        
        self.elements = []
        self.cur_element = None
        
        self.motion_thr = threading.Thread(target = self.motion_thread)
        self.motion_queue = Queue()
        self.motion_thr_running = True        
        self.motion_thr.start()
    
    def stop_threads(self):
        self.motion_queue.put((None, None, None))
        
        self.motion_thr_running = False
        
    def set_fg_color(self, color):
        self.fg_color = color
    
    def set_bg_color(self, color):
        self.bg_color = color
    
    def redraw(self, area = None):
        '''Send expose to parent window'''
        expose_event = gtk.gdk.Event(gtk.gdk.EXPOSE)
        expose_event.window = self.window
        
        if area is not None:
            expose_event.area = area
        else:
            rect = self.get_allocation()
            expose_event.area = gtk.gdk.Rectangle(0, 0, rect.width, rect.height)
        
        self.send_expose(expose_event)
    
    def expose(self, widget, event):
        cr = widget.window.cairo_create() 
        self.rect = self.get_allocation()
        
        surface = cairo.ImageSurface(cairo.FORMAT_RGB24, self.rect.width, self.rect.height)
        scr = cairo.Context(surface)
        
        cr.rectangle(event.area.x, event.area.y, event.area.width, event.area.height)
        cr.clip()
        
        scr.set_source_rgb(1., 1., 1.)
        scr.rectangle(0., 0., self.rect.width, self.rect.height)
        scr.fill()
        
        for el in self.elements:
            el.draw(scr)
            
        cr.set_source(cairo.SurfacePattern(surface))
        cr.paint()
    
    def create_element(self, x, y):
        # Tool to class
        classes = {TOOL_LINE: Line,
                   TOOL_RECTANGLE: Rectange, 
                   TOOL_ELLIPSE: Ellipse, 
                   TOOL_TEXT: Text, 
                   TOOL_SMILE: Smile}
        
        try:
            cls = classes[self.tool]
        except KeyError:
            return False
            
        self.cur_element = cls(x, y, self.fg_color, self.bg_color)
        self.elements.append(self.cur_element)
        
        return True
    
    def on_button_press(self, widget, event):
        if event.button == 1:
            self.create_element(event.x, event.y)
            
            self.start_x = event.x
            self.start_y = event.y
            
            self.redraw(self.cur_element.get_area())
        elif event.button == 3:
            # Remove element 
            new_elements = filter(
                lambda el: not el.inside(event.x, event.y),
                self.elements)
            
            self.elements = new_elements
            
            self.redraw()
    
    def on_motion(self, widget, event):
        if self.cur_element is None:
            return
        
        is_shift = event.state & gtk.gdk.SHIFT_MASK
        self.motion_queue.put((event.x, event.y, is_shift))
    
    def motion_thread(self):
        proc_events = False
        proc_time = time.time()
        
        while self.motion_thr_running:
            if proc_events and (time.time() - proc_time) > 0.01:
                self.async_motion(x, y, is_shift)
                
                gtk.threads_enter()
                # redraw_all needed only for hands
                area = self.cur_element.get_area() \
                    if self.cur_element is not None \
                    else None
                self.redraw(area)
                gtk.threads_leave()
            
                proc_events = False
                proc_time = time.time()
            
            try:
                if self.cur_element is None:
                    # Wait for motion event - polling
                    x, y, is_shift = self.motion_queue.get(False, 0.005)
                else:
                    # Wait for first motion event - blocked
                    x, y, is_shift = self.motion_queue.get()
            except Empty:
                continue
                        
            proc_events = True
    
    def async_motion(self, x, y, is_shift):
        if self.cur_element is not None:
            w, h = x - self.start_x, y - self.start_y
            
            if is_shift:
                w, h = (min(w, h), ) * 2
            
            self.cur_element.update(w, h)
    
    def on_button_release(self, widget, event):
        self.cur_element = None

class TestApplication:
    def delete_event(self, widget, event, data=None):
        return False
    
    def destroy(self, widget, data=None):
        gtk.main_quit()
    
    def create_toolbar(self):
        def create_tb_button(text, file_name, mode, group = True):
            if group:
                group = self.tb_select
            
            return self.toolbar.append_element(gtk.TOOLBAR_CHILD_RADIOBUTTON,
                                               widget = group, text = None,
                                               tooltip_text = text,
                                               tooltip_private_text = None,
                                               icon = gtk.image_new_from_file(file_name),
                                               callback = self.toolbar_callback, 
                                               user_data = mode)
        
        self.toolbar = gtk.Toolbar()
        
        self.tb_select = create_tb_button('Select', 'edit-select.png', TOOL_SELECT, None)
        self.tb_rect = create_tb_button('Dashed line', 'draw-path.png', TOOL_LINE)
        self.tb_rect = create_tb_button('Rectangle', 'draw-rectangle.png', TOOL_RECTANGLE)
        self.tb_ellipse = create_tb_button('Ellipse', 'draw-ellipse.png', TOOL_ELLIPSE)
        self.tb_text = create_tb_button('Draw text', 'draw-text.png', TOOL_TEXT)
        self.tb_smile = create_tb_button('Draw smiley', 'face-smile-big.png', TOOL_SMILE)
        
        self.toolbar.insert_space(-1)
        
        self.toolbar.append_widget(self.fg_color, 'Foreground color', None)
        self.toolbar.append_widget(self.bg_color, 'Background color', None)
        
        self.tb_select.set_active(True)
        
        self.box.pack_start(self.toolbar, False, False)
        
    def toolbar_callback(self, selected, tool, *args):
        if not selected.get_active():
            return
        
        self.drawing_area.tool = tool
       
    def __init__(self):
        self.win = gtk.Window(gtk.WINDOW_TOPLEVEL)
        
        self.win.connect("delete_event", self.delete_event)
        self.win.connect("destroy", self.destroy)
        
        self.win.set_border_width(16)        
        
        self.win.set_geometry_hints(min_width=640, 
                                    min_height=480)
        
        self.fg_color = gtk.ColorButton()
        self.bg_color = gtk.ColorButton()
        self.bg_color.set_color(gtk.gdk.Color(1.0, 1.0, 1.0))
        self.fg_color.connect("color-set", self.color_set)
        self.bg_color.connect("color-set", self.color_set)
        
        self.box = gtk.VBox()
        self.win.add(self.box)        
        
        self.create_toolbar()
        
        self.drawing_area = MyDrawingArea()
        self.box.pack_start(self.drawing_area, True, True)
    
    def color_set(self, colorbtn, *args):
        color = colorbtn.get_color()
        color = (color.red_float, color.green_float, color.blue_float)
        
        if colorbtn == self.fg_color:
            self.drawing_area.set_fg_color(color)
        elif colorbtn == self.bg_color:
            self.drawing_area.set_bg_color(color)
    
    def main(self):
        self.win.show_all()
        
        gtk.threads_init()
        
        try:
            gtk.threads_enter()
            gtk.main()
            gtk.threads_leave()
        finally:
            self.drawing_area.stop_threads()


if __name__ == "__main__":
    app = TestApplication()
    app.main()